@@ -1,9 +1,12 @@
-__version__ = "0.8.0"
+__version__ = "0.9.1"
 
 import os
 
 from .bootstrap import _logging, setup_terrier, setup_jnius, is_windows
 
+# definitive API used by others, now available before pt.init
+from .transformer import Transformer, Estimator, Indexer
+
 import importlib
 
 #sub modules
@@ -22,7 +25,7 @@ pipelines = None
 rewrite = None
 text = None
 transformer = None
-Transformer = None
+
 
 file_path = os.path.dirname(os.path.abspath(__file__))
 firstInit = False
@@ -33,6 +36,7 @@ properties = None
 tqdm = None
 HOME_DIR = None
 init_args ={}
+_helper_version = None
 
 def init(version=None, mem=None, packages=[], jvm_opts=[], redirect_io=True, logging='WARN', home_dir=None, boot_packages=[], tqdm=None, no_download=False,helper_version = None):
     """
@@ -98,11 +102,13 @@ def init(version=None, mem=None, packages=[], jvm_opts=[], redirect_io=True, log
         from os.path import expanduser
         userhome = expanduser("~")
         HOME_DIR = os.path.join(userhome, ".pyterrier")
-        if not os.path.exists(HOME_DIR):
-            os.mkdir(HOME_DIR)
+    if not os.path.exists(HOME_DIR):
+        os.mkdir(HOME_DIR)
 
     # get the initial classpath for the JVM
-    classpathTrJars = setup_terrier(HOME_DIR, version, helper_version = helper_version, boot_packages=boot_packages, force_download=not no_download)
+    classpathTrJars, hlpr_ver = setup_terrier(HOME_DIR, version, helper_version = helper_version, boot_packages=boot_packages, force_download=not no_download)
+    global _helper_version
+    _helper_version = hlpr_ver
     
     if is_windows():
         if "JAVA_HOME" in os.environ:
@@ -133,16 +139,15 @@ def init(version=None, mem=None, packages=[], jvm_opts=[], redirect_io=True, log
     if "BUILD_DATE" in dir(tr_version):
         version_string += " (built by %s on %s)" % (tr_version.BUILD_USER, tr_version.BUILD_DATE)
     import sys
-    print("PyTerrier %s has loaded Terrier %s\n" % (__version__, version_string), file=sys.stderr)
+    print("PyTerrier %s has loaded Terrier %s and terrier-helper %s\n" % (__version__, version_string, _helper_version), file=sys.stderr)
     properties = autoclass('java.util.Properties')()
     ApplicationSetup = autoclass('org.terrier.utility.ApplicationSetup')
 
     from .batchretrieve import BatchRetrieve, FeaturesBatchRetrieve
     from .utils import Utils
     from .datasets import get_dataset, find_datasets, list_datasets
-    from .index import Indexer, FilesIndexer, TRECCollectionIndexer, DFIndexer, DFIndexUtils, IterDictIndexer, FlatJSONDocumentIterator, IndexingType
+    from .index import Indexer, FilesIndexer, TRECCollectionIndexer, DFIndexer, DFIndexUtils, IterDictIndexer, IndexingType, TerrierStemmer, TerrierStopwords, TerrierTokeniser
     from .pipelines import Experiment, GridScan, GridSearch, KFoldGridSearch
-    from .transformer import Transformer
 
     # Make imports global
     globals()["autoclass"] = autoclass
@@ -173,28 +178,34 @@ def init(version=None, mem=None, packages=[], jvm_opts=[], redirect_io=True, log
     _logging(logging)
     setup_jnius()
 
+    # .datasets
     globals()["get_dataset"] = get_dataset
     globals()["list_datasets"] = list_datasets
     globals()["find_datasets"] = find_datasets
-    globals()["Experiment"] = Experiment
+    # .batchretrieve
     globals()["BatchRetrieve"] = BatchRetrieve
     globals()["TerrierRetrieve"] = BatchRetrieve  # TerrierRetrieve is an alias to BatchRetrieve
     globals()["Indexer"] = Indexer
     globals()["FeaturesBatchRetrieve"] = FeaturesBatchRetrieve
+    # .index
     globals()["TRECCollectionIndexer"] = TRECCollectionIndexer
     globals()["FilesIndexer"] = FilesIndexer
     globals()["DFIndexer"] = DFIndexer
     globals()["DFIndexUtils"] = DFIndexUtils
     globals()["IterDictIndexer"] = IterDictIndexer
-    globals()["FlatJSONDocumentIterator"] = FlatJSONDocumentIterator
-    globals()["Utils"] = Utils
     globals()["IndexFactory"] = autoclass("org.terrier.structures.IndexFactory")
     globals()["IndexRef"] = autoclass("org.terrier.querying.IndexRef")
     globals()["IndexingType"] = IndexingType
+    globals()["TerrierStemmer"] = TerrierStemmer
+    globals()["TerrierStopwords"] = TerrierStopwords
+    globals()["TerrierTokeniser"] = TerrierTokeniser
+    # .pipelines etc
+    globals()["Experiment"] = Experiment
     globals()["GridScan"] = GridScan
     globals()["GridSearch"] = GridSearch
     globals()["KFoldGridSearch"] = KFoldGridSearch
-    globals()["Transformer"] = Transformer
+    globals()["Utils"] = Utils
+    
     
     
     # we save the pt.init() arguments so that other processes,
@@ -264,14 +275,14 @@ def version():
     from jnius import autoclass
     return autoclass("org.terrier.Version").VERSION
 
-def check_version(min):
+def check_version(min, helper=False):
     """
         Returns True iff the underlying Terrier version is no older than the requested version.
     """
     from jnius import autoclass
     from packaging.version import Version
     min = Version(str(min))
-    currentVer = Version(version().replace("-SNAPSHOT", ""))
+    currentVer = Version((_helper_version if helper else version()).replace("-SNAPSHOT", ""))
     return currentVer >= min
 
 def redirect_stdouterr():
