@@ -2,13 +2,11 @@ from jnius import autoclass, cast
 from typing import Union
 import pandas as pd
 import numpy as np
-from . import tqdm, check_version
+from . import tqdm, check_version, Transformer
 from warnings import warn
-from .index import Indexer
 from .datasets import Dataset
-from .transformer import TransformerBase, Symbol, is_lambda
+from .transformer import Symbol
 from .model import coerce_queries_dataframe, FIRST_RANK
-import deprecation
 import concurrent
 from concurrent.futures import ThreadPoolExecutor
 
@@ -46,9 +44,10 @@ def _function2wmodel(function):
             #see https://github.com/SeldonIO/alibi/issues/447#issuecomment-881552005
             from dill import extend
             extend(use_dill=False)
-            byterep = pickle.dumps(self.fn)
-            byterep = autoclass("java.nio.ByteBuffer").wrap(byterep)
-            return byterep
+            # keep both python and java representations around to prevent them being GCd in respective VMs 
+            self.pbyterep = pickle.dumps(self.fn)
+            self.jbyterep = autoclass("java.nio.ByteBuffer").wrap(self.pbyterep)
+            return self.jbyterep
 
     callback = PythonWmodelFunction(function)
     wmodel = autoclass("org.terrier.python.CallableWeightingModel")( callback )
@@ -63,13 +62,14 @@ def _mergeDicts(defaults, settings):
 def _parse_index_like(index_location):
     JIR = autoclass('org.terrier.querying.IndexRef')
     JI = autoclass('org.terrier.structures.Index')
+    from .index import TerrierIndexer
 
     if isinstance(index_location, JIR):
         return index_location
     if isinstance(index_location, JI):
         return cast('org.terrier.structures.Index', index_location).getIndexRef()
-    if isinstance(index_location, str) or issubclass(type(index_location), Indexer):
-        if issubclass(type(index_location), Indexer):
+    if isinstance(index_location, str) or issubclass(type(index_location), TerrierIndexer):
+        if issubclass(type(index_location), TerrierIndexer):
             return JIR.of(index_location.path)
         return JIR.of(index_location)
 
@@ -77,10 +77,10 @@ def _parse_index_like(index_location):
         f'''index_location is current a {type(index_location)},
         while it needs to be an Index, an IndexRef, a string that can be
         resolved to an index location (e.g. path/to/index/data.properties),
-        or an pyterrier.Indexer object'''
+        or an pyterrier.index.TerrierIndexer object'''
     )
 
-class BatchRetrieveBase(TransformerBase, Symbol):
+class BatchRetrieveBase(Transformer, Symbol):
     """
     A base class for retrieval
 
@@ -95,7 +95,7 @@ def _from_dataset(dataset : Union[str,Dataset],
             clz,
             variant : str = None, 
             version='latest',            
-            **kwargs):
+            **kwargs) -> Transformer:
 
     from . import get_dataset
     from .io import autoopen
@@ -125,6 +125,22 @@ class BatchRetrieve(BatchRetrieveBase):
     Use this class for retrieval by Terrier
     """
 
+    @staticmethod
+    def matchop(t, w=1):
+        """
+        Static method used for rewriting a query term to use a MatchOp operator if it contains
+        anything except ASCII letters or digits.
+        """
+        import base64
+        import string
+        if not all(a in string.ascii_letters + string.digits for a in t):
+            encoded = base64.b64encode(t.encode('utf-8')).decode("utf-8") 
+            t = f'#base64({encoded})'
+        if w != 1:
+            t = f'#combine:0={w}({t})'
+        return t
+
+
     @staticmethod
     def from_dataset(dataset : Union[str,Dataset], 
             variant : str = None, 
@@ -450,7 +466,7 @@ class BatchRetrieve(BatchRetrieveBase):
 
 
 
-class TextIndexProcessor(TransformerBase):
+class TextIndexProcessor(Transformer):
     '''
         Creates a new MemoryIndex based on the contents of documents passed to it.
         It then creates a new instance of the innerclass and passes the topics to that.
@@ -760,3 +776,74 @@ class FeaturesBatchRetrieve(BatchRetrieve):
         if self.wmodel is None:
             return "FBR(" + str(len(self.features)) + " features)"
         return "FBR(" + self.controls["wmodel"] + " and " + str(len(self.features)) + " features)"
+
+rewrites_setup = False
+
+def setup_rewrites():
+    from .batchretrieve import BatchRetrieve, FeaturesBatchRetrieve
+    from .transformer import rewrite_rules
+    from .ops import FeatureUnionPipeline, ComposedPipeline
+    from matchpy import ReplacementRule, Wildcard, Pattern, CustomConstraint
+    #three arbitrary "things".
+    x = Wildcard.dot('x')
+    xs = Wildcard.plus('xs')
+    y = Wildcard.dot('y')
+    z = Wildcard.dot('z')
+    # two different match retrives
+    _br1 = Wildcard.symbol('_br1', BatchRetrieve)
+    _br2 = Wildcard.symbol('_br2', BatchRetrieve)
+    _fbr = Wildcard.symbol('_fbr', FeaturesBatchRetrieve)
+    
+    # batch retrieves for the same index
+    BR_index_matches = CustomConstraint(lambda _br1, _br2: _br1.indexref == _br2.indexref)
+    BR_FBR_index_matches = CustomConstraint(lambda _br1, _fbr: _br1.indexref == _fbr.indexref)
+    
+    # rewrite nested binary feature unions into one single polyadic feature union
+    rewrite_rules.append(ReplacementRule(
+        Pattern(FeatureUnionPipeline(x, FeatureUnionPipeline(y,z)) ),
+        lambda x, y, z: FeatureUnionPipeline(x,y,z)
+    ))
+    rewrite_rules.append(ReplacementRule(
+        Pattern(FeatureUnionPipeline(FeatureUnionPipeline(x,y), z) ),
+        lambda x, y, z: FeatureUnionPipeline(x,y,z)
+    ))
+    rewrite_rules.append(ReplacementRule(
+        Pattern(FeatureUnionPipeline(FeatureUnionPipeline(x,y), xs) ),
+        lambda x, y, xs: FeatureUnionPipeline(*[x,y]+list(xs))
+    ))
+
+    # rewrite nested binary compose into one single polyadic compose
+    rewrite_rules.append(ReplacementRule(
+        Pattern(ComposedPipeline(x, ComposedPipeline(y,z)) ),
+        lambda x, y, z: ComposedPipeline(x,y,z)
+    ))
+    rewrite_rules.append(ReplacementRule(
+        Pattern(ComposedPipeline(ComposedPipeline(x,y), z) ),
+        lambda x, y, z: ComposedPipeline(x,y,z)
+    ))
+    rewrite_rules.append(ReplacementRule(
+        Pattern(ComposedPipeline(ComposedPipeline(x,y), xs) ),
+        lambda x, y, xs: ComposedPipeline(*[x,y]+list(xs))
+    ))
+
+    # rewrite batch a feature union of BRs into an FBR
+    rewrite_rules.append(ReplacementRule(
+        Pattern(FeatureUnionPipeline(_br1, _br2), BR_index_matches),
+        lambda _br1, _br2: FeaturesBatchRetrieve(_br1.indexref, ["WMODEL:" + _br1.controls["wmodel"], "WMODEL:" + _br2.controls["wmodel"]])
+    ))
+
+    def push_fbr_earlier(_br1, _fbr):
+        #TODO copy more attributes
+        _fbr.wmodel = _br1.controls["wmodel"]
+        return _fbr
+
+    # rewrite a BR followed by a FBR into a FBR
+    rewrite_rules.append(ReplacementRule(
+        Pattern(ComposedPipeline(_br1, _fbr), BR_FBR_index_matches),
+        push_fbr_earlier
+    ))
+
+    global rewrites_setup
+    rewrites_setup = True
+
+setup_rewrites()
\ No newline at end of file
