@@ -2,10 +2,10 @@
 This file contains all the indexers.
 """
 
-# from jnius import autoclass, cast, PythonJavaClass, java_method
 from jnius import autoclass, PythonJavaClass, java_method, cast
+from enum import Enum
 import pandas as pd
-from pyterrier.transformer import IterDictIndexerBase
+from . import Indexer
 import os
 import enum
 import json
@@ -15,8 +15,9 @@ import threading
 import select
 import math
 from warnings import warn
+from deprecated import deprecated
 from collections import deque
-from typing import List, Dict, Union
+from typing import List, Dict, Union, Any
 
 StringReader = None
 HashMap = None
@@ -41,11 +42,17 @@ IndexFactory = None
 StructureMerger = None
 BlockStructureMerger = None
 
+# for backward compatibility
+class IterDictIndexerBase(Indexer):
+    @deprecated(version="0.9", reason="Use pt.Indexer instead of IterDictIndexerBase")
+    def __init__(self, *args, **kwargs):
+        super(Indexer, self).__init__(*args, **kwargs)
 
 # lastdoc ensures that a Document instance from a Collection is not GCd before Java has used it.
 lastdoc=None
 
 def run_autoclass():
+    from . import check_version
     global StringReader
     global HashMap
     global TaggedDocument
@@ -80,7 +87,7 @@ def run_autoclass():
     BlockIndexer = autoclass("org.terrier.structures.indexing.classical.BlockIndexer")
     BasicSinglePassIndexer = autoclass("org.terrier.structures.indexing.singlepass.BasicSinglePassIndexer")
     BlockSinglePassIndexer = autoclass("org.terrier.structures.indexing.singlepass.BlockSinglePassIndexer")
-    BasicMemoryIndexer = autoclass("org.terrier.python.MemoryIndexer")
+    BasicMemoryIndexer = autoclass("org.terrier.realtime.memory.MemoryIndexer" if check_version("5.7") else "org.terrier.python.MemoryIndexer")
     Collection = autoclass("org.terrier.indexing.Collection")
     Arrays = autoclass("java.util.Arrays")
     Array = autoclass('java.lang.reflect.Array')
@@ -245,18 +252,125 @@ def createAsList(files_path : Union[str, List[str]]):
 # Using enum class create enumerations
 class IndexingType(enum.Enum):
     """
-        This enum is used to determine the type of index built by Terrier. The default is CLASSIC.
+        This enum is used to determine the type of index built by Terrier. The default is CLASSIC. For more information,
+        see the relevant Terrier `indexer <https://terrier-core.readthedocs.io/en/latest/indexer_details.html>`_
+        and `realtime <https://terrier-core.readthedocs.io/en/latest/realtime_indices.html>`_ documentation.
     """
     CLASSIC = 1 #: A classical indexing regime, which also creates a direct index structure, useful for query expansion
     SINGLEPASS = 2 #: A single-pass indexing regime, which builds an inverted index directly. No direct index structure is created. Typically is faster than classical indexing.
-    MEMORY = 3 #: An in-memory index. No direct index is created.
+    MEMORY = 3 #: An in-memory index. No persistent index is created.
 
-
-class Indexer:
+class TerrierStemmer(Enum):
     """
-    Parent class. It can be used to load an existing index.
-    Use one of its children classes if you wish to create a new index.
+        This enum provides an API for the stemmers available in Terrier. The stemming configuration is saved in the index
+        and loaded at retrieval time. `Snowball <https://snowballstem.org/>`_ stemmers for various languages 
+        `are available in Terrier <http://terrier.org/docs/current/javadoc/org/terrier/terms/package-summary.html>`_.
+    """
+    none = 'none' #: Apply no stemming
+    porter = 'porter' #: Apply Porter's English stemmer
+    weakporter = 'weakporter' #: Apply a weak version of Porter's English stemmer
+    # available snowball stemmers in Terrier
+    danish = 'danish' #: Snowball Danish stemmer
+    finnish = 'finnish' #: Snowball Finnish stemmer
+    german = 'german' #: Snowball German stemmer
+    hungarian = 'hungarian' #: Snowball Hungarian stemmer
+    norwegian = 'norwegian' #: Snowball Norwegian stemmer
+    portugese = 'portugese' #: Snowball Portuguese stemmer
+    swedish = 'swedish' #: Snowball Swedish stemmer
+    turkish = 'turkish' #: Snowball Turkish stemmer
 
+    @staticmethod
+    def _to_obj(this):
+        try:
+            return TerrierStemmer(this)
+        except ValueError:
+            return this
+
+    @staticmethod
+    def _to_class(this):
+        if this is None or this == TerrierStemmer.none:
+            return None
+        if this == TerrierStemmer.porter:
+            return 'PorterStemmer'
+        if this == TerrierStemmer.weakporter:
+            return 'WeakPorterStemmer'
+        
+        # snowball stemmers
+        if this == TerrierStemmer.danish:
+            return 'DanishSnowballStemmer'
+        if this == TerrierStemmer.finnish:
+            return 'FinnishSnowballStemmer'
+        if this == TerrierStemmer.german:
+            return 'GermanSnowballStemmer'
+        if this == TerrierStemmer.hungarian:
+            return 'HungarianSnowballStemmer'
+        if this == TerrierStemmer.norwegian:
+            return 'NorwegianSnowballStemmer'
+        if this == TerrierStemmer.portugese:
+            return 'PortugueseSnowballStemmer'
+        if this == TerrierStemmer.swedish:
+            return 'SwedishSnowballStemmer'
+        if this == TerrierStemmer.turkish:
+            return 'TurkishSnowballStemmer'
+
+        if isinstance(this, str):
+            return this
+        
+
+class TerrierStopwords(Enum):
+    """
+        This enum provides an API for the stopword configuration used during indexing with Terrier
+    """
+
+    none = 'none' #: No Stemming
+    terrier = 'terrier' #: Apply Terrier's standard stopword list
+
+    @staticmethod
+    def _to_obj(this):
+        try:
+            return TerrierStopwords(this)
+        except ValueError:
+            return this
+
+
+class TerrierTokeniser(Enum):
+    """
+        This enum provides an API for the tokeniser configuration used during indexing with Terrier.
+    """
+
+    whitespace = 'whitespace' #: Tokenise on whitespace only
+    english = 'english' #: Terrier's standard tokeniser, designed for English
+    utf = 'utf' #: A variant of Terrier's standard tokeniser, similar to English, but with UTF support.
+    twitter = 'twitter' #: Like utf, but keeps hashtags etc
+    identity = 'identity' #: Performs no tokenisation - strings are kept as is. 
+
+    @staticmethod
+    def _to_obj(this):
+        try:
+            return TerrierTokeniser(this)
+        except ValueError:
+            return this
+
+    @staticmethod
+    def _to_class(this):
+        if this == TerrierTokeniser.whitespace:
+            return 'WhitespaceTokeniser'
+        if this == TerrierTokeniser.english:
+            return 'EnglishTokeniser'
+        if this == TerrierTokeniser.utf:
+            return 'UTFTokeniser'
+        if this == TerrierTokeniser.twitter:
+            return 'UTFTwitterTokeniser'
+        if this == TerrierTokeniser.identity:
+            return 'IdentityTokeniser'
+        if isinstance(this, str):
+            return this
+
+class TerrierIndexer:
+    """
+    This is the super class for all of the Terrier-based indexers exposed by PyTerrier. It hosts common configuration
+    for all index types.
+    
     """
 
     default_properties = {
@@ -267,16 +381,29 @@ class Indexer:
             "trec.collection.class": "TRECCollection",
     }
 
-    def __init__(self, index_path, *args, blocks=False, overwrite=False, verbose=False, meta_reverse=["docno"], type=IndexingType.CLASSIC, **kwargs):
+    def __init__(self, index_path : str, *args, 
+            blocks : bool = False, 
+            overwrite: bool = False, 
+            verbose : bool = False, 
+            meta_reverse : List[str] = ["docno"],
+            stemmer : Union[None, str, TerrierStemmer] = TerrierStemmer.porter,
+            stopwords : Union[None, TerrierStopwords] = TerrierStopwords.terrier,
+            tokeniser : Union[str,TerrierTokeniser] = TerrierTokeniser.english,
+            type=IndexingType.CLASSIC, 
+            **kwargs):
         """
-        Init method
+        Constructor called by all indexer subclasses. All arguments listed below are available in 
+        IterDictIndexer, DFIndexer, TRECCollectionIndexer and FilesIndsexer. 
 
         Args:
             index_path (str): Directory to store index. Ignored for IndexingType.MEMORY.
             blocks (bool): Create indexer with blocks if true, else without blocks. Default is False.
             overwrite (bool): If index already present at `index_path`, True would overwrite it, False throws an Exception. Default is False.
             verbose (bool): Provide progess bars if possible. Default is False.
-            type (IndexingType): the specific indexing procedure to use. Default is IndexingType.CLASSIC.
+            stemmer (TerrierStemmer): the stemmer to apply. Default is ``TerrierStemmer.porter``.
+            stopwords (TerrierStopwords): the stopwords list to apply. Default is ``TerrierStemmer.terrier``.
+            tokeniser (TerrierTokeniser): the stemmer to apply. Default is ``TerrierTokeniser.english``.
+            type (IndexingType): the specific indexing procedure to use. Default is ``IndexingType.CLASSIC``.
         """
         if StringReader is None:
             run_autoclass()
@@ -290,6 +417,9 @@ class Indexer:
         self.index_dir = index_path
         self.blocks = blocks
         self.type = type
+        self.stemmer = TerrierStemmer._to_obj(stemmer)
+        self.stopwords = TerrierStopwords._to_obj(stopwords)
+        self.tokeniser = TerrierTokeniser._to_obj(tokeniser)
         self.properties = Properties()
         self.setProperties(**self.default_properties)
         self.overwrite = overwrite
@@ -336,12 +466,11 @@ class Indexer:
 
     def createIndexer(self):
         """
-        TODO
         Checks `self.type` and
         - if false, check `blocks` and create a BlockIndexer if true, else create BasicIndexer
         - if true, check `blocks` and create a BlockSinglePassIndexer if true, else create BasicSinglePassIndexer
         Returns:
-            Created index object
+            Created  Java object extending org.terrier.structures.indexing.Indexer
         """
         
         Indexer, _ = self.indexerAndMergerClasses()
@@ -360,10 +489,37 @@ class Indexer:
         Returns:
             type objects for indexer and merger for the given configuration
         """
+
+        # configure the meta index
         self.properties['indexer.meta.forward.keys'] = ','.join(self.meta.keys())
         self.properties['indexer.meta.forward.keylens'] = ','.join([str(l) for l in self.meta.values()])
         self.properties['indexer.meta.reverse.keys'] = ','.join(self.meta_reverse)
+
+        # configure the term pipeline
+        if 'termpipelines' in self.properties:
+            # use existing configuration if present
+            warn("Setting of termpipelines property directly is deprecated", stacklevel=4, category=DeprecationWarning)
+        else:
+            
+            termpipeline = []
+            if self.stopwords is not None and self.stopwords != TerrierStopwords.none:
+                termpipeline.append('Stopwords')
+
+            stemmer_clz = TerrierStemmer._to_class(self.stemmer)
+            if stemmer_clz is not None:
+                termpipeline.append(stemmer_clz)
+            
+            self.properties['termpipelines'] = ','.join(termpipeline)
+
+        if "tokeniser" in self.properties:
+            warn("Setting of tokeniser property directly is deprecated", stacklevel=4, category=DeprecationWarning)
+        else:
+            self.properties['tokeniser'] = TerrierTokeniser._to_class(self.tokeniser)
+
+        # inform terrier of all properties
         ApplicationSetup.getProperties().putAll(self.properties)
+
+        # now create the indexers 
         if self.type is IndexingType.SINGLEPASS:
             if self.blocks:
                 Indexer = BlockSinglePassIndexer
@@ -479,7 +635,7 @@ class DFIndexUtils:
                 ),
             lengths)
 
-class DFIndexer(Indexer):
+class DFIndexer(TerrierIndexer):
     """
     Use this Indexer if you wish to index a pandas.Dataframe
 
@@ -594,9 +750,8 @@ class FlatJSONDocumentIterator(PythonJavaClass):
             return lastdoc
         return None
 
-from pyterrier.transformer import IterDictIndexerBase
-class _BaseIterDictIndexer(Indexer, IterDictIndexerBase):
-    def __init__(self, index_path, *args, meta = {'docno' : 20}, meta_reverse=['docno'], threads=1, **kwargs):
+class _BaseIterDictIndexer(TerrierIndexer, Indexer):
+    def __init__(self, index_path, *args, meta = {'docno' : 20}, meta_reverse=['docno'], pretokenised=False, threads=1, **kwargs):
         """
         
         Args:
@@ -604,11 +759,18 @@ class _BaseIterDictIndexer(Indexer, IterDictIndexerBase):
             meta(Dict[str,int]): What metadata for each document to record in the index, and what length to reserve. Defaults to `{"docno" : 20}`.
             meta_reverse(List[str]): What metadata shoudl we be able to resolve back to a docid. Defaults to `["docno"]`,      
         """
-        IterDictIndexerBase.__init__(self)
-        Indexer.__init__(self, index_path, *args, **kwargs)
+        Indexer.__init__(self)
+        TerrierIndexer.__init__(self, index_path, *args, **kwargs)
         self.threads = threads
         self.meta = meta
         self.meta_reverse = meta_reverse
+        self.pretokenised = pretokenised
+        if self.pretokenised:
+            from pyterrier import check_version
+            assert check_version(5.7), "Terrier too old, this requires 5.7"
+            # we disable stemming and stopwords for pretokenised indices
+            self.stemmer = None
+            self.stopwords = None
 
     def _setup(self, fields, meta, meta_lengths):
         """
@@ -630,15 +792,26 @@ class _BaseIterDictIndexer(Indexer, IterDictIndexerBase):
                 meta_lengths = ['512'] * len(meta)
             self.meta = { k:v for k,v in zip( meta, meta_lengths)}
 
-        self.setProperties(**{
-            'metaindex.compressed.crop.long' : 'true',
-            'FieldTags.process': ','.join(fields),
-            'FieldTags.casesensitive': 'true',
-        })
+        if self.pretokenised:
+            self.setProperties(**{
+                'metaindex.compressed.crop.long' : 'true',
+                'FieldTags.process': '',
+                'FieldTags.casesensitive': 'true',
+            })
+        else:
+            self.setProperties(**{
+                'metaindex.compressed.crop.long' : 'true',
+                'FieldTags.process': ','.join(fields),
+                'FieldTags.casesensitive': 'true',
+            })
 
     def _filter_iterable(self, it, indexed_fields):
         # Only include necessary fields: those that are indexed, metadata fields, and docno
-        all_fields = {'docno'} | set(indexed_fields) | set(self.meta.keys())
+        
+        if self.pretokenised:
+            all_fields = {'docno', "toks"} | set(self.meta.keys())
+        else:
+            all_fields = {'docno'} | set(indexed_fields) | set(self.meta.keys())
         for doc in it:
             yield {f: doc[f] for f in all_fields}
 
@@ -666,15 +839,35 @@ class _IterDictIndexer_nofifo(_BaseIterDictIndexer):
 
         self._setup(fields, self.meta, None)
         assert self.threads == 1, 'IterDictIndexer does not support multiple threads on Windows'
-        # we need to prevent collectionIterator from being GCd
-        collectionIterator = FlatJSONDocumentIterator(self._filter_iterable(it, fields))
-        javaDocCollection = autoclass("org.terrier.python.CollectionFromDocumentIterator")(collectionIterator)
+
         index = self.createIndexer()
-        index.index([javaDocCollection])
-        global lastdoc
-        lastdoc = None
-        self.index_called = True
-        collectionIterator = None
+        if self.pretokenised:
+            assert not self.blocks, "pretokenised isnt compatible with blocks"
+
+            # we generate DocumentPostingList from a dictionary of pretokenised text, e.g.
+            # [
+            #     {'docno' : 'd1', 'toks' : {'a' : 1, 'aa' : 2}}
+            # ]
+            
+            iter_docs = DocListIterator(it)
+            self.index_called = True
+            index.indexDocuments(iter_docs)
+            iter_docs = None
+
+        else:
+
+            # we need to prevent collectionIterator from being GCd
+            collectionIterator = FlatJSONDocumentIterator(self._filter_iterable(it, fields))
+            javaDocCollection = autoclass("org.terrier.python.CollectionFromDocumentIterator")(collectionIterator)
+            # remove once 5.7 is now the minimum version
+            from . import check_version
+            index.index(javaDocCollection if check_version("5.7") else [javaDocCollection])
+            global lastdoc
+            lastdoc = None
+            self.index_called = True
+            collectionIterator = None
+
+
         if self.type is IndexingType.MEMORY:
             return index.getIndex().getIndexRef()
         return IndexRef.of(self.index_dir + "/data.properties")
@@ -698,6 +891,8 @@ class _IterDictIndexer_fifo(_BaseIterDictIndexer):
         """
         CollectionFromDocumentIterator = autoclass("org.terrier.python.CollectionFromDocumentIterator")
         JsonlDocumentIterator = autoclass("org.terrier.python.JsonlDocumentIterator")
+        if self.pretokenised:
+            JsonlTokenisedIterator = autoclass("org.terrier.python.JsonlPretokenisedIterator")
         ParallelIndexer = autoclass("org.terrier.python.ParallelIndexer")
 
         if meta is not None:
@@ -726,7 +921,10 @@ class _IterDictIndexer_fifo(_BaseIterDictIndexer):
             for i in range(self.threads):
                 fifo = f'{d}/docs-{i}.jsonl'
                 os.mkfifo(fifo)
-                j_collections.append(CollectionFromDocumentIterator(JsonlDocumentIterator(fifo)))
+                if (self.pretokenised):
+                    j_collections.append(JsonlTokenisedIterator(fifo))
+                else:
+                    j_collections.append(CollectionFromDocumentIterator(JsonlDocumentIterator(fifo)))
                 fifos.append(fifo)
 
             # Start dishing out the docs to the fifos
@@ -735,11 +933,17 @@ class _IterDictIndexer_fifo(_BaseIterDictIndexer):
             # Different process for memory indexer (still taking advantage of faster fifos)
             if Indexer is BasicMemoryIndexer:
                 index = Indexer()
-                index.index(j_collections)
+                if self.pretokenised:
+                    index.indexDocuments(j_collections)
+                else:
+                    index.index(j_collections)
                 return index.getIndex().getIndexRef()
 
             # Start the indexing threads
-            ParallelIndexer.buildParallel(j_collections, self.index_dir, Indexer, Merger)
+            if self.pretokenised:
+                ParallelIndexer.buildParallelTokenised(j_collections, self.index_dir, Indexer, Merger)
+            else:
+                ParallelIndexer.buildParallel(j_collections, self.index_dir, Indexer, Merger)
             return IndexRef.of(self.index_dir + "/data.properties")
 
     def _write_fifos(self, it, fifos):
@@ -761,6 +965,8 @@ class _IterDictIndexer_fifo(_BaseIterDictIndexer):
                         # single threaded mode
                         ready = deque(fifos)
                 fifo = ready.popleft()
+                if 'toks' in doc:
+                    doc['toks'] = {k: int(v) for k, v in doc['toks'].items()} # cast all values as ints
                 json.dump(doc, fifo)
                 fifo.write('\n')
 
@@ -774,7 +980,7 @@ else:
 IterDictIndexer.__name__ = 'IterDictIndexer' # trick sphinx into not using "alias of"
 
 
-class TRECCollectionIndexer(Indexer):
+class TRECCollectionIndexer(TerrierIndexer):
 
 
     """
@@ -783,14 +989,12 @@ class TRECCollectionIndexer(Indexer):
 
     def __init__(self, 
             index_path : str, 
-            blocks : bool = False, 
-            overwrite : bool = False, 
-            type : IndexingType =IndexingType.CLASSIC, 
             collection : str = "trec", 
             verbose : bool = False,
             meta : Dict[str,int] = {"docno" : 20},
             meta_reverse : List[str] = ["docno"],
-            meta_tags : Dict[str,str] = {}
+            meta_tags : Dict[str,str] = {},
+            **kwargs
             ):
         """
         Init method
@@ -806,7 +1010,7 @@ class TRECCollectionIndexer(Indexer):
             meta_tags(Dict[str,str]): For collections formed using tagged data (e.g. HTML), which tags correspond to which metadata. This is useful for recording the text of documents for use in neural rankers - see :ref:`pt.text`.
 
         """
-        super().__init__(index_path, blocks=blocks, overwrite=overwrite, type=type)
+        super().__init__(index_path, **kwargs)
         if isinstance(collection, str):
             if collection in type_to_class:
                 collection = type_to_class[collection]
@@ -831,12 +1035,15 @@ class TRECCollectionIndexer(Indexer):
         _TaggedDocumentSetup(self.meta, self.meta_tags)
 
         colObj = createCollection(files_path, self.collection)
-        collsArray = [colObj]
         if self.verbose and isinstance(colObj, autoclass("org.terrier.indexing.MultiDocumentFileCollection")):
             colObj = cast("org.terrier.indexing.MultiDocumentFileCollection", colObj)
             colObj = TQDMCollection(colObj)
-            collsArray = autoclass("org.terrier.python.PTUtils").makeCollection(colObj)
-        index.index(collsArray)
+        import pyterrier as pt
+        # remove once 5.7 is now the minimum version
+        if pt.check_version("5.7"):
+            index.index(colObj)
+        else:
+            index.index(autoclass("org.terrier.python.PTUtils").makeCollection(colObj))
         global lastdoc
         lastdoc = None
         colObj.close()
@@ -845,7 +1052,7 @@ class TRECCollectionIndexer(Indexer):
             return index.getIndex().getIndexRef()
         return IndexRef.of(self.index_dir + "/data.properties")
 
-class FilesIndexer(Indexer):
+class FilesIndexer(TerrierIndexer):
     '''
         Use this Indexer if you wish to index a pdf, docx, txt etc files
 
@@ -879,7 +1086,9 @@ class FilesIndexer(Indexer):
         _FileDocumentSetup(self.meta, self.meta_tags)
         
         simpleColl = SimpleFileCollection(asList, False)
-        index.index([simpleColl])
+        # remove once 5.7 is now the minimum version
+        from . import check_version
+        index.index(simpleColl if check_version("5.7") else [simpleColl])
         global lastdoc
         lastdoc = None
         self.index_called = True
@@ -922,7 +1131,59 @@ class TQDMSizeCollection(PythonJavaClass):
         lastdoc = self.collection.getDocument()
         return lastdoc
         
+class DocListIterator(PythonJavaClass):
+    dpl_class = autoclass("org.terrier.structures.indexing.DocumentPostingList")
+    tuple_class = autoclass("org.terrier.structures.collections.MapEntry")
+    __javainterfaces__ = [
+        'java/util/Iterator',
+    ]
 
+    def __init__(self, pyiterator):
+        self.pyiterator = pyiterator
+        self.hasnext = True
+        self.lastdoc = None
+        import pyterrier as pt
+        self.tr57 = not pt.check_version("5.8")
+
+    @staticmethod 
+    def pyDictToMap(a_dict): #returns Map<String,String>
+        rtr = HashMap()
+        for k,v in a_dict.items():
+            rtr.put(k, v)
+        return rtr
+
+    def pyDictToMapEntry(self,doc_dict : Dict[str,Any]): #returns Map.Entry<Map<String,String>, DocumentPostingList>>
+        dpl = DocListIterator.dpl_class()
+        # this works around a bug in the counting of doc lengths in Tr 5.7
+        if self.tr57:
+            for t,tf in doc_dict["toks"].items():
+                for i in range(int(tf)):
+                    dpl.insert(t)
+        else: #this code for 5.8 onwards
+            for t,tf in doc_dict["toks"].items():
+                for i in range(int(tf)):
+                    dpl.insert(int(tf), t)
+        
+        # we cant make the toks column into the metaindex as it isnt a string. remove it.
+        del doc_dict["toks"]
+        return DocListIterator.tuple_class(DocListIterator.pyDictToMap(doc_dict), dpl)
+
+    @java_method('()Z')
+    def hasNext(self):
+        return self.hasnext
+
+    @java_method('()Ljava/lang/Object;')
+    def next(self):
+        try:
+            doc_dict = next(self.pyiterator)
+        except StopIteration as se:
+            self.hasnext = False
+            # terrier will ignore a null return from an iterator
+            return None
+        # keep this around to prevent being GCd before Java can read it
+        self.lastdoc = self.pyDictToMapEntry(doc_dict)
+        return self.lastdoc
+        
 
 class TQDMCollection(PythonJavaClass):
     __javainterfaces__ = ['org/terrier/indexing/Collection']
