@@ -2,10 +2,12 @@ import pyterrier as pt
 from jnius import cast
 import pandas as pd
 from .batchretrieve import _parse_index_like
-from .transformer import TransformerBase
+from . import Transformer
+from .index import TerrierTokeniser
 from . import tqdm
 from warnings import warn
-from typing import List
+from typing import List,Union
+from types import FunctionType
 
 TerrierQLParser = pt.autoclass("org.terrier.querying.TerrierQLParser")()
 TerrierQLToMatchingQueryTerms = pt.autoclass("org.terrier.querying.TerrierQLToMatchingQueryTerms")()
@@ -30,7 +32,65 @@ def _check_terrier_prf():
             break
     assert _terrier_prf_package_loaded, _terrier_prf_message
 
-def reset() -> TransformerBase:
+def tokenise(tokeniser : Union[str,TerrierTokeniser,FunctionType] = 'english', matchop=False) -> Transformer:
+    """
+
+    Applies tokenisation to the query. By default, queries obtained from ``pt.get_dataset().get_topics()`` are
+    normally tokenised.
+
+    Args:
+        tokeniser(Union[str,TerrierTokeniser,FunctionType]): Defines what tokeniser should be used - either a Java tokeniser name in Terrier, a TerrierTokeniser instance, or a function that takes a str as input and returns a list of str.
+        matchop(bool): Whether query terms should be wrapped in matchops, to ensure they can be parsed by a Terrier BatchRetrieve transformer.
+    
+    Example - use default tokeniser::
+
+        pipe = pt.rewrite.tokenise() >> pt.BatchRetrieve()
+        pipe.search("Question with 'capitals' and other stuff?")
+    
+    Example - roll your own tokeniser::
+
+        poortokenisation = pt.rewrite.tokenise(lambda query: query.split(" ")) >> pt.BatchRetrieve()
+
+    Example - for non-English languages, tokenise on standard UTF non-alphanumeric characters::
+
+        utftokenised = pt.rewrite.tokenise(pt.TerrierTokeniser.utf)) >> pt.BatchRetrieve()
+        utftokenised = pt.rewrite.tokenise("utf")) >> pt.BatchRetrieve()
+
+    Example - tokenising queries using a `HuggingFace tokenizer <https://huggingface.co/docs/transformers/fast_tokenizers>`_ ::
+
+        # this assumes the index was created in a pretokenised manner
+        br = pt.BatchRetrieve(indexref)
+        tok = AutoTokenizer.from_pretrained("bert-base-uncased")
+        query_toks = pt.rewrite.tokenise(tok.tokenize, matchop=True)
+        retr_pipe = query_toks >> br
+    
+    """
+    _query_fn = None
+    if isinstance(tokeniser, FunctionType):
+        _query_fn = tokeniser
+    else:
+        tokeniser = TerrierTokeniser._to_obj(tokeniser)
+        tokeniser = TerrierTokeniser._to_class(tokeniser)
+        if "." not in tokeniser:
+            tokeniser = 'org.terrier.indexing.tokenisation.' + tokeniser
+        tokenobj = pt.autoclass(tokeniser)()
+        _query_fn = tokenobj.getTokens
+
+    def _join_str(input : Union[str,List[str]]):
+        if isinstance(input, str):
+            return input
+        return ' '.join(input)
+
+    def _join_str_matchop(input : List[str]):
+        assert not isinstance(input, str), "Expected a list of strings"
+        return ' '.join(map(pt.BatchRetrieve.matchop, input))
+    
+    if matchop:
+        return pt.apply.query(lambda r: _join_str_matchop(_query_fn(r.query)))
+    return pt.apply.query(lambda r: _join_str(_query_fn(r.query)))
+
+
+def reset() -> Transformer:
     """
         Undoes a previous query rewriting operation. This results in the query formulation stored in the `"query_0"`
         attribute being moved to the `"query"` attribute, and, if present, the `"query_1"` being moved to
@@ -49,7 +109,7 @@ def reset() -> TransformerBase:
     from .model import pop_queries
     return pt.apply.generic(lambda topics: pop_queries(topics))
 
-class SDM(TransformerBase):
+class SDM(Transformer):
     '''
         Implements the sequential dependence model, which Terrier supports using its
         Indri/Galagoo compatible matchop query language. The rewritten query is derived using
@@ -68,6 +128,9 @@ class SDM(TransformerBase):
         assert check_version("5.3")
         self.ApplyTermPipeline_stopsonly = pt.autoclass("org.terrier.querying.ApplyTermPipeline")("Stopwords")
 
+    def __repr__(self):
+        return "SDM()"
+
     def transform(self, topics_and_res):
         results = []
         from .model import ranked_documents_to_queries, push_queries
@@ -122,7 +185,7 @@ class SequentialDependence(SDM):
     '''
     pass
     
-class QueryExpansion(TransformerBase):
+class QueryExpansion(Transformer):
     '''
         A base class for applying different types of query expansion using Terrier's classes.
         This transformer changes the query. It must be followed by a Terrier Retrieve() transformer.
@@ -184,6 +247,14 @@ class QueryExpansion(TransformerBase):
             raise ValueError("Input resultset has neither docid nor docno")
         return QueryResultSet(docids, scores, occurrences)
 
+    def __repr__(self):
+        return "QueryExpansion(" + ",".join([
+            self.indexref.toString(),
+            str(self.fb_docs),
+            str(self.fb_terms),
+            str(self.qe)
+            ]) + ")"
+
     def _configure_request(self, rq):
         rq.setControl("qe_fb_docs", str(self.fb_docs))
         rq.setControl("qe_fb_terms", str(self.fb_terms))
@@ -357,7 +428,7 @@ class AxiomaticQE(QueryExpansion):
         self.qe.fbDocs = self.fb_docs
         return super().transform(queries_and_docs)
 
-def stash_results(clear=True) -> TransformerBase:
+def stash_results(clear=True) -> Transformer:
     """
     Stashes (saves) the current retrieved documents for each query into the column `"stashed_results_0"`.
     This means that they can be restored later by using `pt.rewrite.reset_results()`.
@@ -369,14 +440,14 @@ def stash_results(clear=True) -> TransformerBase:
     """
     return _StashResults(clear)
     
-def reset_results() -> TransformerBase:
+def reset_results() -> Transformer:
     """
     Applies a transformer that undoes a `pt.rewrite.stash_results()` transformer, thereby restoring the
     ranked documents.
     """
     return _ResetResults()
 
-class _StashResults(TransformerBase):
+class _StashResults(Transformer):
 
     def __init__(self, clear, *args, **kwargs):
         super().__init__(*args, **kwargs)
@@ -407,9 +478,12 @@ class _StashResults(TransformerBase):
                 for i in range(len(groupDf)):
                     groupDf.at[i, "stashed_results_0"] = docsDict
                 rtr.append(groupDf)
-            return pd.concat(rtr)        
+            return pd.concat(rtr)   
 
-class _ResetResults(TransformerBase):
+    def __repr__(self):
+        return "pt.rewrite.stash_results()"     
+
+class _ResetResults(Transformer):
 
     def transform(self, topics_with_saved_docs : pd.DataFrame) -> pd.DataFrame:
         if "stashed_results_0" not in topics_with_saved_docs.columns:
@@ -426,7 +500,10 @@ class _ResetResults(TransformerBase):
             rtr.append(finaldf)
         return pd.concat(rtr)
 
-def linear(weightCurrent : float, weightPrevious : float, format="terrierql", **kwargs) -> TransformerBase:
+    def __repr__(self):
+        return "pt.rewrite.reset_results()"
+
+def linear(weightCurrent : float, weightPrevious : float, format="terrierql", **kwargs) -> Transformer:
     """
     Applied to make a linear combination of the current and previous query formulation. The implementation
     is tied to the underlying query language used by the retrieval/re-ranker transformers. Two of Terrier's
@@ -453,7 +530,7 @@ def linear(weightCurrent : float, weightPrevious : float, format="terrierql", **
     """
     return _LinearRewriteMix([weightCurrent, weightPrevious], format, **kwargs)
 
-class _LinearRewriteMix(TransformerBase):
+class _LinearRewriteMix(Transformer):
 
     def __init__(self, weights : List[float], format : str = 'terrierql', **kwargs):
         super().__init__(**kwargs)
@@ -488,3 +565,6 @@ class _LinearRewriteMix(TransformerBase):
         newDF = push_queries(topics_and_res)
         newDF["query"] = newDF.apply(fn, axis=1)
         return newDF
+
+    def __repr__(self):
+        return "pt.rewrite.linear()"
\ No newline at end of file
